/*
  libservo360b.c  ver 2017/12/8
                    This is the test harness / demo for the servo360b library
       Uses a Spin/PASM driver by Tom Crawford
       This revision controls and reports rotational velocity
       
     Group sections in this order:
      Demo global declarations
      Demo functions
      main()
      
*/




#include "simpletools.h"                      // Include simple tools
#include "stdint.h"
//#include "servo360b.h"

// ---------------------------------------------
//  This starts the code for servo360.h
// ---------------------------------------------
int MyParm[15];               // passes addresses of parameters
volatile int8_t mode[4], Complete[4];
volatile int16_t ServoPins[4];         //  must be [4], ControlPin<<8 | FeedbackPin
volatile int16_t PulseWidth[4], theta[4], commanded[4], index ;
volatile int Context[4], Context2[4], Parameters[4], Velocity[4]; 
volatile int *cog360servo;

// ***  servo names
#define Servo0 0
#define Servo1 1
#define Servo2 2
#define Servo3 3
#define LWheel 0
#define RWheel 1

// **** DRIVER function prototypes
void start360(int16_t *, int16_t *);        // CPIN, FPIN arrays
void SetPos(short, short);          // theservo, commanded[theservo] in degrees
void Width(short, short);           // theservo, microsecs, command servo with pulse width
short GetPos(short);                // theservo, returns theta in degrees
int8_t CheckDone(short);            // theservo, returns Complete 0 if at position, !=0 if not

void SetVelocity(short, int, int);  // theservo, RPM (10-100), Direction(0=cw,1=ccw) 
short GetPulseWidth(short);   // new fn: theservo, returns commanded PulseWidth 
int GetVelocity(short);       // new fn: theservo, rtn actual velocity*3000/4096 = RPM
                              //                    negative is CCW
void fbsdrive(int, int);      // RPML, RPMR -- left and right servos + = fwd, - = rev

void SetMaxError(short, int8_t);    // theservo, error +1 in degrees
void SetGain(short, int8_t);        // theservo, gainP 
void SetDeltaMin(short, int8_t);    // theservo, Dmin  
void SetDeltaMax(short, int8_t);    // theservo, Dmax 
// ---------------------------------------------
//  This ends the code for servo360.h
// ---------------------------------------------

// ---------------------------------------------
//  This starts the code for servo360.c
// ---------------------------------------------

/*
void SetVelocity(short, int, int);  // theservo, RPM (10-100), Direction(0=cw,1=ccw) 
short GetPulseWidth(short);   // new fn: theservo, returns commanded PulseWidth 
int GetVelocity(short);       // new fn: theservo, rtn actual velocity*3000/4096 = RPM
                             //                    negative is CCW 
*/
//#include "servo360a.h"

//  **** Declare DRIVER globals in servo360a.h

//  ======================================================
//            START DRIVER FUNCTIONS
//  ======================================================

//  ***************************
//      Initialize
//  ***************************
void start360(short *CPIN, short *FPIN)     // pin numbers from define in header
{
 ServoPins[0] = CPIN[0] << 8 | FPIN[0];
 ServoPins[1] = CPIN[1] << 8 | FPIN[1];
 ServoPins[2] = CPIN[2] << 8 | FPIN[2];
 ServoPins[3] = CPIN[3] << 8 | FPIN[3];

 for( int i = 0; i <= 3; i++)
   {
     if (ServoPins[i] != 0)
      {
        mode[i] = 0;
        PulseWidth[i] = 1500;                         
        Parameters[i] = (3<<24)|(18 <<16)|(31 <<8)|150 ;    
        // error-3, gain-12, deltamin-31, deltamax-150
      } // end if
    }   // end for i
 MyParm[0] = PulseWidth;
 MyParm[1] = mode; 
 MyParm[2] = theta;
 MyParm[3] = commanded; 
 MyParm[4] = Complete;        
 MyParm[5] = CLKFREQ / 1000000;
 MyParm[6] = ServoPins; 
 MyParm[7] = Context;
 MyParm[8] = Parameters;
 MyParm[9] = Context2;
 MyParm[10] = Velocity;

 extern int binary_FeedBackServo4_dat_start[]; 
 cog360servo = cognew((void*)binary_FeedBackServo4_dat_start, MyParm); 
 pause(500);          
 }
 
//  ******************************************************
//        Order servo to position - degrees 
//  ******************************************************
void SetPos(short servonum, short degrees)
{
  commanded[servonum] = degrees;
  Complete[servonum] = 1;                  //we are not complete yet
  mode[servonum] = 1;                      //set position mode
 }
 
//  ******************************************************
//      Make servo receive pulses of spec'd width
//  ****************************************************** 
void Width(short servonum, short Usec)
{
  mode[servonum]= 0;                //  CPW mode
  PulseWidth[servonum] = Usec;      // save pulse width in usec
 }
 
//  ******************************************************
//     Set velocity in RPM (10 to 100)
//    directions, CW = 0, CCW = 1
//  ******************************************************  
void SetVelocity(short servonum, int rpm, int direction) 
{
  mode[servonum] = 0;
  PulseWidth[servonum] = 1500;
  commanded[servonum] = (rpm *4096) / 3000;   // convert to clicks per tic 4096 steps/rev,3000 pulses/min 
 if(direction != 0) commanded[servonum] =  -1 * commanded[servonum]; // CCW
  mode[servonum] = 8;     // set velocity mode
} 
 
//  ******************************************************
//      Get the current position of the servo
//  ****************************************************** 
short GetPos(short servonum)
{
  int16_t xxx = theta[servonum];
  return xxx;
  //return theta[servonum];    // can't return an array element
}  

//  ******************************************************
//     Check if servo is at commanded position
//     Returns Complete[servonum]
//       =0 if in commanded position +/- error for 200ms
//      !=0 if not in position 
//  ******************************************************  
int8_t CheckDone(short servonum)
{
  int8_t xxx = Complete[servonum];
  return xxx;
  //return Complete[servonum];    // can't return an array element
}  

//  ******************************************************
//      Get the current pulse width of the servo 
//        Use in Velocity mode
//  ****************************************************** 
short GetPulseWidth(short servonum)
{
  int16_t xxx = PulseWidth[servonum];
  return xxx; 
}

//  ******************************************************
//      Get the current velocity (RPM) of the servo
//         returns RPM[servo] + = CW, - = CCW
//  ****************************************************** 
int GetVelocity(short servonum)
{
  int xxx = Velocity[servonum];
  return xxx;
}  

//  ******************************************************
//    Define the maximum allowable error for spec'd servo
//      accepterror = 1 degree greater than acceptable error
//  ****************************************************** 
void SetMaxError(short servonum, int8_t accepterror)
{
  Parameters[servonum] = Parameters[servonum] & 0x00FFFFFF;
  Parameters[servonum] = Parameters[servonum] | (accepterror << 24);  
}

//  ******************************************************
//   Define the gain for the specified servo. 
//    typical 5 to 15, higher results in more overshoot   
//  ******************************************************
void SetGain(short servonum, int8_t gainP)
{
  Parameters[servonum] = Parameters[servonum] & 0xFF00FFFF;  
  Parameters[servonum] = Parameters[servonum] | (gainP << 16);
}

//  ******************************************************
//   Specifies the minimum deviation from 1500 usec 
//    Use to keep servo from being stuck in dead zone
//    Typical values approx 30  
//  ******************************************************
void SetDeltaMin(short servonum, int8_t Dmin) 
{
  Parameters[servonum] = Parameters[servonum] & 0xFFFF00FF;  
  Parameters[servonum] = Parameters[servonum] | (Dmin << 8);  
}

//  ******************************************************
//    Specifies the maximum deviation from 1500 usec.
//      Typical value approx 150
//  ******************************************************
void SetDeltaMax(short servonum, int8_t Dmax)
{
  Parameters[servonum] = Parameters[servonum] & 0xFFFFFF00;  
  Parameters[servonum] = Parameters[servonum] | Dmax;   
}

//  ******************************************************
//    ActivityBot drive command
//      Parameters: Left Wheel RPM, Right Wheel RPM
//                  For both + = forward, - = reverse --- 
//                  Function converts to +/- for each wheel
//    cases:  Command inL inR    outL     outR
//            forward  +   +    CCW = 1  CW  = 0
//            reverse  -   -    CW  = 0  CCW = 1
//            Turn-R   +   -    CCW = 1  CCW = 1
//            Turn-L   -   +    CW  = 0  CW  = 0
//  ******************************************************
void fbsdrive(int RPML, int RPMR)
{
  // left wheel velocity SetVelocity(short servonum, int rpm, int direction)
  // LWheel = servo0
   int Ldir = 1, Rdir = 1, Lval, Rval;
   if(RPML < 0) Ldir = 0; //SetVelocity(ii, RPML, 0);
  // right wheel velocity SetVelocity(short servonum, int rpm, int direction)
  // RWheel = servo1  
   if(RPMR > 0) Rdir = 0;  //SetVelocity(ii, RPMR, 0);
   Lval = abs(RPML);
   Rval = abs(RPMR);
   SetVelocity(0, Lval, Ldir);
   SetVelocity(1, Rval, Rdir);
}

//  ======================================================
//        END DRIVER FUNCTIONS 
//  ======================================================
// ---------------------------------------------
//  This ends the code for servo360.c
// ---------------------------------------------


//  ======================================================
//       DEMO defines, variables, function prototypes
//  ======================================================
// THESE DEFINES AND DECLARATIONS NEED TO BE IN ANY PROGRAM THAT USES
//   THE servo360a.h HEADER
// **** defines for demo
#define CP0 12      // also for ActivityBot left wheel (CCW = fwd)
#define CP1 13      // also for ActivityBot right wheel (CW = fwd) 
#define CP2 14
#define CP3 0
#define FP0 8
#define FP1 9
#define FP2 10 
#define FP3 0
 int16_t CPIN[] = {CP0, CP1, CP2, CP3};
 int16_t FPIN[] = {FP0, FP1, FP2, FP3};
//  ** END mandatory declares & defines

// **** Declare demo globals
// ***  Servo pins control and feedback, if both 0 -- no servo

volatile int8_t  deltaMax, deltaMin, Cog, DisplayServo, currentservo, testgain;  
volatile int16_t ordered, actual, rcount, theservo, Cmnded[4];    // short sub ordered for commanded
int random;

// **** Demo function prototypes
void MySet(short, short);       //  servonum, degrees
short WaitSet(short);            //  servonum,  returns position in deg
void showangles();

//  ======================================================
//        Start DEMO FUNCTIONS 
//  ======================================================
//  ******************************************************
//         Set ordered position
//  ******************************************************
void MySet(short servonum, short degrees)
{
  SetPos(servonum, degrees);
}   

//  ******************************************************
//    Wait until servo has completed ordered movement     
//  ******************************************************
short WaitSet(short servonum)
{
  while(CheckDone(servonum));
  pause(10);
  return GetPos(servonum);
}

//  ******************************************************
//    Print servo position in degrees     
//  ******************************************************
void showangles()
{
  short servonum;
  for(servonum = 0; servonum <= 2; servonum++)
    {
       print("  Servo%d %d  ",servonum, GetPos(servonum));
    }        
}
//  ******************************************************
//        END DEMO FUNCTIONS 
//  ******************************************************
//  ******************************************************

int main()                   // The DEMO
{
 start360(CPIN, FPIN); 
 print("cog360 = %d \n", cog360servo);
 pause(500);

// Parameters values are set for all servos in library function start360()to
// error-3, gain-18, deltamin-31, deltamax-150 
// Set parameters to fit your servos below 

// ***** Change parameters for individual servos 
 SetMaxError(Servo0, 4);
SetMaxError(Servo1, 4); 
SetMaxError(Servo2, 4);
 pause(10);
 
 SetGain(Servo0, 16); // 22
 SetGain(Servo1, 12); // 24
 //SetGain(Servo2, 20);
  pause(10); 
 
 SetDeltaMin(Servo0, 30);  //29
 SetDeltaMin(Servo1, 30); // 29
 SetDeltaMin(Servo2, 31);
 pause(10);
//  ***** End parameter set *****

//  ***** command servos to move *****
 Width(Servo0, 1450);   // start servos rotating 1280 - 1720
 Width(Servo1, 1650);
 Width(Servo2, 1450);
 Width(Servo3, 1550);
 
 for(int i = 0; i < 50; i++)
  {
    showangles();
    print("\n");
    pause(10);
  }       
 for(theservo = 0; theservo <= 2; theservo++) Width(theservo,1500);   // stop rotation
 pause(2000);

// rotate servo0 by two degrees
  int degree = 1;
  MySet(0, degree);
  actual =  WaitSet(0);
  print("\nCommanded: %3d  Actual: %3d", degree, actual);   
  for(int i = 2; i<=90; i=i+2)
    {
      MySet(0, i);
      actual =  WaitSet(0);
      print("\nCommanded: %3d  Actual: %3d  Error: %3d", i, actual, i-actual);
      pause(100);
    } 
  
// Large scale movements for all servos
 for(theservo = 0; theservo <= 2; theservo++)   // full scale changes
  {
    print("\nservo %d  ",theservo);
    MySet(theservo, 180);
    actual = WaitSet(theservo);
    print("%d  ",actual);
    
    MySet(theservo, 0);
    actual = WaitSet(theservo);
    print("%d  ",actual); 
       
    MySet(theservo, 358);
    actual = WaitSet(theservo);    
    print("%d  ",actual); 
       
    MySet(theservo, 0);
    actual = WaitSet(theservo);
    print("%d  \n",actual);
  }                             // end full scale changes


// Move servos to random positions
  srand(CNT);
  random = rand();
  rcount = 0;
  while(rcount < 50)
    {
      for(theservo = 0; theservo <= 2; theservo++)   // 
      {
       if(CheckDone(theservo) == 0)
        {
         print("\n%d  Commanded: %3d  Actual: %3d", theservo, Cmnded[theservo], GetPos(theservo));
         print("  Error: %2d", Cmnded[theservo] - GetPos(theservo));
         
         random = rand();
         ordered = abs(random) % 360;
         MySet(theservo, ordered);
         pause(200);
         Cmnded[theservo] = ordered;
         print("   %2d", rcount++);
        }   // end if      
      }     // end for 
    }       // end while
       

// ***********************************************
//  Control Robot Wheel Speeds  fbsdrive(L-RPM, R-RPM)
// ***********************************************

//  try 20 RPM fwd straight 
  print("fwd 20 RPM\n");
  fbsdrive(20, 20);
  pause(3000);    // do for 3 seconds

//  speed up to 60 RPM straight 
  print("fwd 60 RPM\n");
  fbsdrive(60, 60);
  pause(3000);    // do for 3 seconds
  
//  try stop both -- FbsDrive(0,0)
  print(" 0 RPM\n");
  fbsdrive(0, 0);
  pause(3000);    // do for 3 seconds

//  reverse 20 rpm both 
  print("rev 20 RPM\n");
  fbsdrive(-20, -20);
  pause(3000);    // do for 3 seconds
  fbsdrive(0, 0); 
  pause(500);  

// try pivot right  FbsDrive(+10,-10)
  print("Pivot Rt\n");
  fbsdrive(20, -20);
  pause(3000);    // do for 3 seconds
  fbsdrive(0, 0);  
  pause(500);

// now pivot left FbsDrive (-10,+10)
  print("Pivot left\n");
  fbsdrive(-20, 20);
  pause(3000);    // do for 3 seconds
  fbsdrive(0, 0);  
  pause(500);   

// straighten out FbsDrive(+20,+20)
  print("fwd 20 RPM\n");
  fbsdrive(20, 20);
  pause(3000);    // do for 3 seconds

// and stop FbsDrive(0,0)
  fbsdrive(0, 0);
  pause(500);   
    
  SetVelocity(0, 40, 0);
  pause(3000);    // do for 3 seconds
  fbsdrive(0, 0);  
  pause(500);

  SetVelocity(0, 40, 1);
  pause(3000);    // do for 3 seconds
  fbsdrive(0, 0);  
  pause(500);
    
}   // end main

